'''
Test how certain variables influence simulation result.

Public Functions:

Test influence of one variable:
- test_var1:    Test how a certain patch variable (mu, w, kappa) influences results.
                Run simulations across different values of that variable and save data.
- var_UV1:      Plot how U, V change across different values of a specified variable.
- var_pi1:      Plot how Upi, V_pi change across different values of a specified variable.
                    

Test influence of two variables:
- test_var2:    Test how two patch variables influence results.
                Run simulations across different values of the two specified variables and save data.
- var_UV2:      Plot how U, V change across the two variables.
                x-axis is values of the variable 2, y-axis is U or V. Values of variable 1 are shown by multiple curves.
- var_pi2:      Plot how Upi & V_pi change across the two variables.


Additional tools (can be called directly):
- get_dirs1:   test_var1 returns directories in a certain format based on var and values. This function mimics that format and 
                        returns a 1D list of directories, where all data are stored.
                        This 1D list is intended to pass to var_UV1 and var_pi1 directly.
    
- var_convergence1: Find the simulatoin results of test_var1 that diverge.

- get_dirs2:   test_var2 returns directories in a certain format based on vars and values. This function mimics that format and
                    returns a 2D list of directories, where all data are stored.
                    This 2D list is intended to pass to var_UV2 and var_pi2 directly.
- var_convergence2: Find the simulatoin results of test_var2 that diverge.
'''

from . import simulation
from .tools import figure_tools as figure_t
from . import analysis as analysis
from . import data_tools as data_t

import matplotlib.pyplot as plt
import numpy as np



# curve type used by var_UV, var2_UV, var_pi, var2_pi
# can be 'o-', 'x-', ...
DOTTED_CURVE_TYPE = 'o-'

# map patch_var name to index in the patch class (in simulation.py)
PATCH_VAR_DICT = {'mu1': 0, 'mu2': 1, 'w1': 2, 'w2': 3, 'kappa1': 4, 'kappa2': 5}

# display name (in latex)
PATCH_VAR_DISP = {'mu1': r'$\mu_U$', 'mu2': r'$\mu_V$', 'w1': r'$w_U$', 'w2': r'$w_V$', 
                  'kappa1': r'$\kappa_U$', 'kappa2': r'$\kappa_V$'}



def test_var1(mod, var, values, dirs, compress_itv = None):
    '''
    Test the influence of one patch variable on simulation results.

    Inputs::
    - sim:                a simulation.model object. All tests will use parameters of mod, except for the variable to test.
    - var:                str, which patch variable to test. e.g. var can be 'mu1', 'w2', 'kappa2', ...
    - values:             1D np.array or list, what values of var to test.
    - dirs:               str, where to save data.
    - compress_itv:         int, whether to reduce data size (if not 1), passed to simulation.model.compress_data function.
    - predict_runtime:    bool, whether to predict how much time left for each test.          

    Returns::
    - var_dirs:           a 1D list of directories (str) where all the data are stored. 
                        Directories have form 'mu1=0.1'. 
                        NOTE: if you accidentally lost this return value, you can retrieve it by get_dirs1.
    '''

    if var not in PATCH_VAR_DICT.keys():
        raise ValueError('Please enter a valid patch variable: mu1, mu2, w1, w2, kappa1, or kappa2')

    # var_dirs is where data will be stored
    var_dirs = []
    
    for k in range(len(values)):
        sim2 = mod.copy(copy_data = False)
        current_var_str = var + '=' + str(values[k])    # e.g., 'mu1=0.1'
        var_dirs.append(dirs + '/' + current_var_str)
        
        for i in range(mod.N):
            for j in range(mod.M):
                sim2.P[i][j][PATCH_VAR_DICT[var]] = values[k]

        try:
            simulation.run(sim2, message = current_var_str + ', ')
            if compress_itv != None:
                sim2.compress_data(compress_itv)
            data_t.save_data(sim2, var_dirs[k], print_msg = False)
        except OverflowError:
            print(current_var_str + ' raised OverflowError, skipped')
            continue

    return var_dirs



def test_var2(mod, var1, var2, values1, values2, dirs, compress_itv = None):
    '''
    Two-variable version of test_var1. Test the influence of two varibles on simulation results.

    Inputs:
    - sim:      a simulation.model object. All tests will use the parameters of mod, except for the two vars to be tested.
    - var1:     str, the first variable to test.
    - var2:     str, the second variable to test.
    - values1:  1D list or np.array, values for var1.
    - values2:  1D list or np.array, values for var2.
    - dirs, compress_itv, scale_maxtime, predict_runtime: same as in test_var1

    Returns:
    - var_dirs:   2D list of directories, where all the data are stored.
                Directories have form 'mu1=0.1, mu2=0.2'.
                NOTE: if you accidentally lost this return value, you can retrieve by get_dirs2.
    '''

    if (var1 not in PATCH_VAR_DICT.keys()) or (var2 not in PATCH_VAR_DICT.keys()):
        raise ValueError('Please enter a valid patch variable: mu1, mu2, w1, w2, kappa1, or kappa2')

    var_dirs = [[] for k1 in range(len(values1))]

    for k1 in range(len(values1)):
        for k2 in range(len(values2)):
            sim2 = mod.copy(copy_data = False)
            current_var_str = var1 + '=' + str(values1[k1]) + ', ' + var2 + '=' + str(values2[k2])   # e.g., mu1=0.1, mu2=0.2
            var_dirs[k1].append(dirs + '/' + current_var_str)

            for i in range(mod.N):
                for j in range(mod.M):
                    sim2.P[i][j][PATCH_VAR_DICT[var1]] = values1[k1]
                    sim2.P[i][j][PATCH_VAR_DICT[var2]] = values2[k2]

            try:
                simulation.run(sim2, message = current_var_str + ', ')
                if compress_itv != None:
                    sim2.compress_data(compress_itv)
                data_t.save_data(sim2, var_dirs[k1][k2], print_msg = False)
            except OverflowError:
                print(current_var_str + ' raise OverflowError, skipped')
                continue

    return var_dirs




def var_UV1(var, values, var_dirs, ax_U = None, ax_V = None, start = 0.95, end = 1.0, U_color = 'purple', V_color = 'green'):
    '''
    Plot function for test_var1, plot how var influences U, V population.
    Make U, V - var curve in two figures, with y-axis being total population at the end of simulations, x-axis being var's values.

    Inputs:
    - var:        str, which variable was tested.
    - values:     1D list or np.array, which values were tested.
    - var_dirs:   return value of test_var1, a 1D list of directories where all data are stored.
    - ax_U, ax_V:   matplotlib axes to plot on.
    - start, end: floats, give an interval of time over which to take average and make plot.
                For example, start = 0.95, end = 1.0 are to take average over the last 5% time of results; 
                essentially plots how var influences equilibrium population.
    
    Returns:
    - ax_U, ax_V: axes for U, V, respectively.
    '''

    # average value of U, V over the interval. One entry for one value of var
    U_ave = []
    V_ave = []

    for k in range(len(var_dirs)):
        try:
            simk = data_t.read_data(var_dirs[k])
        except FileNotFoundError:
            print(var + '=' + str(values[k]) + ' not found, skipped')
            U_ave.append(None)
            V_ave.append(None)
            continue
        start_index = int(simk.max_record * start)
        end_index = int(simk.max_record * end)
        NM = int(simk.N * simk.M)

        U_ave.append(sum(figure_t.ave_interval_1D(simk.U, start_index, end_index)) / NM)
        V_ave.append(sum(figure_t.ave_interval_1D(simk.V, start_index, end_index)) / NM)
        
    #### plot ####
    if ax_U == None:
        _, ax_U = plt.subplots()
    if ax_V == None:
        _, ax_V = plt.subplots()
    var_disp = PATCH_VAR_DISP[var]
    
    ax_U.set_xlabel(var_disp)
    ax_U.set_ylabel('Population ' + r'$U$')
    ax_U.plot(values, U_ave, DOTTED_CURVE_TYPE, color = U_color)
    ax_U.set_title(figure_t.gen_title(var_disp + r'$\,-\,U$', start, end))
    
    ax_V.set_xlabel(var_disp)
    ax_V.set_ylabel('Population ' + r'$V$')
    ax_V.plot(values, V_ave, DOTTED_CURVE_TYPE, color = V_color)
    ax_V.set_title(figure_t.gen_title(var_disp + r'$\,-\,V$', start, end))
    
    return ax_U, ax_V




def var_UV2(var1, var2, values1, values2, var_dirs, ax_U = None, ax_V = None, var1_on_xaxis = True, start = 0.95, end = 1.0, U_color = 'viridis', V_color = 'viridis', rgb_alpha = 1):
    '''
    Plot function for test_var2, plot how two variables influence U, V population.
    Make U, V - var1, var2 curves. y-axis is population, x-axis is var2's values, 
    and var1's values are represented by different curves, one curve corresponds to one value of var1.

    Inputs:
    - var1:       str, the first variable tested.
    - var2:       str, the second variable tested.
    - values1:    1D list or np.array, values for var1.
    - values2:    1D list or np.array, values for var2.
    - var_dirs:   return value of test_var2, a 2D list of directories where all data are stored.
    - ax_U, ax_V:   matplotlib axes to plot on.
    - var1_on_xaxis:    whether to put var1 on x-axis. Set to False if not.
    - start, end: floats, give an interval of time over which to take average and make plot.
                For example, start = 0.95, end = 1.0 plots the near-end population (equilibrium, if converged).
    - color:      str, what colors to use for different value of var1. Uses Matplotlib color maps. 
                See available color maps at: https://matplotlib.org/stable/users/explain/colors/colormaps.html
    - rgb_alpha:  the alpha value for color. Thr curves might have overlaps, recommend set a small alpha value if so.
    
    Returns:
    - ax_U, ax_V: axes for U, V, respectively.
    '''

    # average value of U, V over the interval. One entry for one values of var1, var2
    U_ave = [[] for k1 in range(len(var_dirs))]
    V_ave = [[] for k1 in range(len(var_dirs))]

    for k1 in range(len(var_dirs)):
        for k2 in range(len(var_dirs[k1])):
            try:
                simk = data_t.read_data(var_dirs[k1][k2])
            except FileNotFoundError:
                print(var1 + '=' + str(values1[k1]) + ', ' + var2 + '=' + str(values2[k2]) + ' not found, skipped')
                U_ave[k1].append(None)
                V_ave[k1].append(None)
                continue
            start_index = int(simk.max_record * start)
            end_index = int(simk.max_record * end)
            NM = int(simk.N * simk.M)

            U_ave[k1].append(sum(figure_t.ave_interval_1D(simk.U, start_index, end_index)) / NM)
            V_ave[k1].append(sum(figure_t.ave_interval_1D(simk.V, start_index, end_index)) / NM)

    U_ave = np.array(U_ave)
    V_ave = np.array(V_ave)
    
    #### plot ####
    if ax_U == None:
        _, ax_U = plt.subplots()
    if ax_V == None:
        _, ax_V = plt.subplots()

    var1_disp = PATCH_VAR_DISP[var1]
    var2_disp = PATCH_VAR_DISP[var2]
    

    cmap_U = plt.get_cmap(U_color)
    if var1_on_xaxis:
        for k in range(len(values2)):
            label = var2_disp + '=' + str(values2[k])
            color_k = cmap_U(k / len(values2))[:3] + (rgb_alpha,)
            ax_U.plot(values1, U_ave[:, k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_U.set_xlabel(var1_disp)
    else:
        for k in range(len(values1)):
            label = var1_disp + '=' + str(values1[k])
            color_k = cmap_U(k / len(values1))[:3] + (rgb_alpha,)
            ax_U.plot(values2, U_ave[k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_U.set_xlabel(var2_disp)
    ax_U.set_ylabel('Population ' + r'$U$')
    ax_U.set_title(figure_t.gen_title(var1_disp + ',' + var2_disp + r'$\,-\,U$', start, end))
    ax_U.legend()
    
    cmap_V = plt.get_cmap(V_color)
    if var1_on_xaxis:
        for k in range(len(values2)):
            label = var2_disp + '=' + str(values2[k])
            color_k = cmap_V(k / len(values2))[:3] + (rgb_alpha,)
            ax_V.plot(values1, V_ave[:, k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_V.set_xlabel(var1_disp)
    else:
        for k in range(len(values1)):
            label = var1_disp + '=' + str(values1[k])
            color_k = cmap_V(k / len(values1))[:3] + (rgb_alpha,)
            ax_V.plot(values2, V_ave[k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_V.set_xlabel(var2_disp)
    ax_V.set_ylabel('Population ' + r'$V$')
    ax_V.set_title(figure_t.gen_title(var1_disp + ',' + var2_disp + r'$\,-\,V$', start, end))
    ax_V.legend()
    
    return ax_U, ax_V



def var_pi1(var, values, var_dirs, ax_U = None, ax_V = None, start = 0.95, end = 1.0, U_color = 'violet', V_color  = 'yellowgreen'):
    '''
    Plot function for test_var1. Plot influence of var on U, V's payoff.
    Make Upi, V_pi - var curves. y-axis is payoff, x-axis is values of var.

    Inputs:
    - var_dirs:   return value of test_var1. A 1D list of directories where all data are stored.
    - var:        str, which variable as tested.
    - values:     1D list or np.array, what values were used.
    - ax_U, ax_V:   matplotlib axes to plot on.
    - start, end: floats, define an interval of time over which to calculate average payoff and make plots.
    
    Returns:
    - ax_U, ax_V: axes for U, V, respectively.
    '''
    
    # take average value of payoff over an interval of time
    U_ave = []
    V_ave = []
    
    for k in range(len(var_dirs)):
        try:
            simk = data_t.read_data(var_dirs[k])
        except FileNotFoundError:
            print(var + '=' + str(values[k]) + ' not found, skipped')
            U_ave.append(None)
            V_ave.append(None)
            continue
        start_index = int(simk.max_record * start)
        end_index = int(simk.max_record * end)
        NM = int(simk.N * simk.M)

        U_ave.append(np.sum(figure_t.ave_interval(simk.Upi, start_index, end_index)) / NM)
        V_ave.append(np.sum(figure_t.ave_interval(simk.Vpi, start_index, end_index)) / NM)

        del simk
            
    #### plot ####
    if ax_U == None:
        _, ax_U = plt.subplots()
    if ax_V == None:
        _, ax_V = plt.subplots()
    var_disp = PATCH_VAR_DISP[var]
    
    ax_U.set_xlabel(var_disp)
    ax_U.set_ylabel('Payoff ' + r'$p_U$')
    ax_U.plot(values, U_ave, DOTTED_CURVE_TYPE, color = U_color)
    ax_U.set_title(figure_t.gen_title(var_disp + r'$\,-\,p_U$', start, end))
    
    ax_V.set_xlabel(var_disp)
    ax_V.set_ylabel('Payoff ' + r'$p_V$')
    ax_V.plot(values, V_ave, DOTTED_CURVE_TYPE, color = V_color)
    ax_V.set_title(figure_t.gen_title(var_disp  + r'$\,-\,p_V$', start, end))

    return ax_U, ax_V




def var_pi2(var1, var2, values1, values2, var_dirs, ax_U = None, ax_V = None, var1_on_xaxis = True, start = 0.95, end = 1.0, U_color = 'viridis', V_color = 'viridis', rgb_alpha = 1):
    '''
    Plot function for test_var2. Plot how var1 and var2 influence payoff.
    Make Upi, V_pi - var2 curves. y-axis is payoff, x-axis is values of var2,
    var1 is represented by different curves. One curve corresponds to one value of var1.

    Inputs:
    - var_dirs:           return value of test_var2. A 2D list of directories where all data are stored.
    - var1, var2:         str, what variables were tested.
    - values1, values2:   1D lists or np.array, what values were tested.
    - ax_U, ax_V:           matplotlib axes to plot on.
    - var1_on_xaxis:    whether to put var1 on x-axis. Set to False if not.
    - start, end:         floats, define a time inteval over which to make average and make plots.
    - color:              str, Matplotlib color maps.
    - rgb_alpha:          set alpha value for curves.
    
    Returns:
    - ax_U, ax_V:         U, V's payoff figures, respectively.
    '''
    
    # take average value of payoff over an interval of time
    U_ave = [[] for k1 in range(len(var_dirs))]
    V_ave = [[] for k1 in range(len(var_dirs))]

    for k1 in range(len(var_dirs)):
        for k2 in range(len(var_dirs[k1])):
            try:
                simk = data_t.read_data(var_dirs[k1][k2])
            except FileNotFoundError:
                print(var1 + '=' + str(values1[k1]) + ', ' + var2 + '=' + str(values2[k2]) + ' not found, skipped')
                U_ave[k1].append(None)
                V_ave[k1].append(None)
                continue
            start_index = int(simk.max_record * start)
            end_index = int(simk.max_record * end)
            NM = int(simk.N * simk.M)

            U_ave[k1].append(np.sum(figure_t.ave_interval(simk.Upi, start_index, end_index)) / NM)
            V_ave[k1].append(np.sum(figure_t.ave_interval(simk.Vpi, start_index, end_index)) / NM)

            del simk    # manually delete this large object

    U_ave = np.array(U_ave)
    V_ave = np.array(V_ave)
    
    #### plot ####
    if ax_U == None:
        _, ax_U = plt.subplots()
    if ax_V == None:
        _, ax_V = plt.subplots()

    var1_disp = PATCH_VAR_DISP[var1]
    var2_disp = PATCH_VAR_DISP[var2]
    
    cmap_U = plt.get_cmap(U_color)
    if var1_on_xaxis:
        for k in range(len(values2)):
            label = var2_disp + '=' + str(values2[k])
            color_k = cmap_U(k / len(values2))[:3] + (rgb_alpha,)
            ax_U.plot(values1, U_ave[:, k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_U.set_xlabel(var1_disp)
    else:
        for k in range(len(values1)):
            label = var1_disp + '=' + str(values1[k])
            color_k = cmap_U(k / len(values1))[:3] + (rgb_alpha,)
            ax_U.plot(values2, U_ave[k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_U.set_xlabel(var2_disp)
    ax_U.set_ylabel('Payoff ' + r'$p_{V}$')
    ax_U.set_title(figure_t.gen_title(var1_disp + ',' + var2_disp + r'$\,-\,p_U$', start, end))
    ax_U.legend()
    
    cmap_V = plt.get_cmap(V_color)
    if var1_on_xaxis:
        for k in range(len(values2)):
            label = var2_disp + '=' + str(values2[k])
            color_k = cmap_V(k / len(values2))[:3] + (rgb_alpha,)
            ax_V.plot(values1, V_ave[:, k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_V.set_xlabel(var1_disp)
    else:
        for k in range(len(values1)):
            label = var1_disp + '=' + str(values1[k])
            color_k = cmap_V(k / len(values1))[:3] + (rgb_alpha,)
            ax_V.plot(values2, V_ave[k], DOTTED_CURVE_TYPE, label = label, color = color_k)
            ax_V.set_xlabel(var2_disp)
    ax_V.set_ylabel('Payoff ' + r'$p_{U}$')
    ax_V.set_title(figure_t.gen_title(var1_disp + ',' + var2_disp + r'$\,-\,p_V$', start, end))
    ax_V.legend()
    
    return ax_U, ax_V




def get_dirs1(var, values, dirs):
    '''
    Mimics the return value format of test_var1 and returns a 1D list of directories where test_var1 saved data.
    Intended usage: retrieve the directories if you accidentally lost the return value of test_var1.

    Inputs:
    - var:        what variable was tested.
    - values:     what values were testsed.
    - dirs:  the directory you passed to test_var1 as 'dirs' parameter. 
                Essentially the root directories where data were stored.

    Returns:
    - var_dirs:   a 1D list of directories where data were stored. Has the same format as the return value of test_var1.
    '''

    var_dirs = []
    values_sorted = sorted(values)
    if dirs[-1] != '/':
        dirs += '/'

    for val in values_sorted:
            var_dirs.append(dirs + var + '=' + str(val))

    return var_dirs




def var_convergence1(var_dirs, interval = 20, start = 0.8, fluc = 0.07):
    '''
    Find the simulatoin results of test_var1 that diverge. 

    Inputs:
    - var_dirs:       Return value of test_var1
    - interval:       int. One of the inputs of analysis.check_convergence.
                    The size of interval to take average over.
    - start:          (0,1) float. One of the inputs of analysis.check_convergence. 
                    Convergence is expected to start from at least this point.
    - fluc:           (0,1) float. One of the inputs of analysis.check_convergence. 
                    Expect the difference between any two small intervals (a quotient-form difference) should be less than fluc.

    Returns:
    - diverge_list:   A list directories where the simulation data diverge.
    '''

    diverge_list = []

    for dirs in var_dirs:
        try:
            mod = data_t.read_data(dirs)
        except FileNotFoundError:
            print(dirs + ' data not found, skipped')
            continue
        if not analysis.check_convergence(mod, interval, start, fluc):
            diverge_list.append(dirs)

    return diverge_list




def get_dirs2(var1, var2, values1, values2, dirs):
    '''
    Mimics the return value format of test_var2 and returns a 2D list of directories where test_var2 saved data.
    Intended usage: retrieve the directories if you accidentally lost the return value of test_var2.

    Inputs:
    - var1, var2:         what variables were tested.
    - values1, values2:   what values were testsed.
    - dirs:          the directory you passed to test_var2 as 'dirs' parameter. 
                        Essentially the root directories where data were stored.

    Returns:
    - var_dirs:   a 2D list of directories where data were stored. Has the same format as the return value of test_var2.
    '''

    var_dirs = [[] for i in range(len(values1))]
    values1_sorted = sorted(values1)
    values2_sorted = sorted(values2)
    if dirs[-1] != '/':
        dirs += '/'

    for i in range(len(values1)):
        for j in range(len(values2)):
            v1 = values1_sorted[i]
            v2 = values2_sorted[j]
            dirs_ij = dirs + var1 + '=' + str(v1) + ', ' + var2 + '=' + str(v2)
            var_dirs[i].append(dirs_ij)

    return var_dirs




def var_convergence2(var_dirs, interval = 20, start = 0.8, fluc = 0.07):
    '''
    Find the simulatoin results of test_var2 that diverge.

    Inputs:
    - var_dirs:       Return value of test_var2
    - interval:       int. One of the inputs of analysis.check_convergence.
                    The size of interval to take average over.
    - start:          (0,1) float. One of the inputs of analysis.check_convergence. 
                    Convergence is expected to start from at least this point.
    - fluc:           (0,1) float. One of the inputs of analysis.check_convergence. 
                    Expect the difference between any two small intervals (a quotient-form difference) should be less than fluc.

    Returns:
    - diverge_list:   A list directories where the simulation data diverge.
    '''

    diverge_list = []

    for sublist in var_dirs:
        for dirs in sublist:
            try:
                mod = data_t.read_data(dirs)
            except FileNotFoundError:
                print(dirs + ' data not found, skipped')
                continue
            if not analysis.check_convergence(mod, interval, start, fluc):
                diverge_list.append(dirs)

    return diverge_list






__version__ = '2.2.2'

'''
version history:

0.1.0: first publishing, May 11, 2025
0.1.1: fix dependency errors
0.1.2: fixing module not find error
0.1.3: restructuring package
0.1.4 ~ 0.1.6: fixing moviepy import issue
0.1.7: changed name back to 'piegy'
0.1.8: updated installation in README
0.1.9: first round of full debugging

1.0.0: first version in PyPI.
1.1.0: debugging. Updated a range of functions, in the following modules: figures, videos, test_var, model, figure_tools
1.1.1: minor debugging in model module.
1.1.2: fix text bad location in figure_tools, update labeling and titling in figures and test_var. Add dpi param to make_video in videos. Remove reset_data function in model.
1.1.3: update README.
1.1.4: changed name: ``model`` module to ``simulation``, and ``model.simulation`` class to ``simulation.model``. Bug fix in videos.
1.1.5: update README.
1.1.6: change name of variables in model class -- for compatability with the new C core. 1.1.6 is the last verion of v1. From v2 on, the piegy package has C core.

2.0.0: update simulation core to C-based.
2.0.1: re-upload, the C core is not included in package.
2.0.2: update version number in __version__.py and update README.
2.0.3: speed boost & debugging in C core. Add check_overflow feature in simulation module, checks whether overflow/too large numbers might be encountered in simulation.
2.0.4: minor debuggings.
2.0.5: fix error in random number generator.
2.1.0: redo random number generator. Update package upload so that more compatible across platforms.
2.1.1: fix import bug for the C core.
2.1.2 ~ 2.1.9: updating & fixing wheel.
2.1.10: fix print bug in run function.
2.1.11: fix .so duplicate error.
2.2.1: change heatmap plotting tool from Seaborn to Matplotlib. Change video maker to cv2 (opencv-python).
2.2.2: impose stricter overflow error check. Used update_mig_one function (already written before, didn't use). This gives 30% speed improvement.
'''


'''
Make mp4 videos for simulation results.

Videos are made by:
make every frame by figures.py functions, then put frames together into a video.

Public Function:
- make_video:   make video based simulation results.

Private Functions
- get_max_lim:  Get the max lim (interval) over many lims, and then expand it a bit for better accommodation.
                Essentially takes union of those intervals. 
- video_lim:    Find a large enough xlim and ylim for video.
- make_mp4:     Put frames together into a mp4.
others not documented here.

'''


from . import figures
from .tools import file_tools as file_t
from .tools import figure_tools as figure_t

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.style as mplstyle
import numpy as np
import os
from cv2 import imread, VideoWriter, VideoWriter_fourcc
from moviepy import ImageSequenceClip


# a list of supported figures
SUPPORTED_FIGURES = ['UV_hmap', 'pi_hmap', 'UV_bar', 'pi_bar', 'UV_hist', 'pi_hist', 'UV_pi']


# map function name to functios in figures.py
# functions not in this dictionary is not supported for videos.
FUNC_DICT = {'UV_hmap': figures.UV_hmap, 'UV_bar': figures.UV_bar, 'UV_hist': figures.UV_hist, 
             'pi_hmap': figures.pi_hmap, 'pi_bar': figures.pi_bar, 'pi_hist': figures.pi_hist, 'UV_pi': figures.UV_pi}


# Map some color maps to regular colors, used to change colors when an invalid color name is given
SNS_PLT_COLOR_DICT = {'Greens': 'green', 'Purples': 'purple', 'BuPu': 'violet', 'YlGn': 'yellowgreen'}
# Map regular colors to color maps
PLT_SNS_COLOR_DICT = {'green': 'Greens', 'purple': 'Purples', 'violet': 'BuPu', 'yellowgreen': 'YlGn'}




def convert_color(func_name, U_color, V_color):
    '''
    Converts some invalid colors.
    If making heatmap videos but gave single colors, map to color maps.
    If making barplot or histogram videos but gave single colors, map to Matplotlib
    '''

    if 'hmap' in func_name:
        # if making heatmaps but give regular colors
        if U_color in PLT_SNS_COLOR_DICT.keys():
            print('Making heatmaps, changed \'' + U_color + '\' to \'' + PLT_SNS_COLOR_DICT[U_color] + '\'')
            U_color = PLT_SNS_COLOR_DICT[U_color]
        if V_color in PLT_SNS_COLOR_DICT.keys():
            print('Making heatmaps, changed \'' + V_color + '\' to \'' + PLT_SNS_COLOR_DICT[V_color] + '\'')
            V_color = PLT_SNS_COLOR_DICT[V_color]
        
        return U_color, V_color

    elif 'hmap' not in func_name:
        # if making barplots or histogram
        if U_color in SNS_PLT_COLOR_DICT.keys():
            print('Not making heatmaps, changed \'' + U_color + '\' to \'' + SNS_PLT_COLOR_DICT[U_color] + '\'')
            U_color = SNS_PLT_COLOR_DICT[U_color]
        if V_color in SNS_PLT_COLOR_DICT.keys():
            print('Not making heatmaps, changed \'' + V_color + '\' to \'' + SNS_PLT_COLOR_DICT[V_color] + '\'')
            V_color = SNS_PLT_COLOR_DICT[V_color]

        return U_color, V_color



def get_max_lim(lims):
    '''
    Get the max lim over many lims, i.e., the lowest lower bound and highest upper bound.
    And then expand it a bit for better accommodation.

    Input:
        lim:    list or np.array, has form [lim1, lim2, ...]
    
    Returns:
        A max lim which contains all lims.
    '''

    lims = np.array(lims)
    
    lim_min = np.min(lims[:, 0]) # min of min
    lim_max = np.max(lims[:, 1]) # max of max
    r = lim_max - lim_min

    if lim_min != 0:
        # negative values are reached
        # extend both upper bound and lower bound 
        return [lim_min - r * 0.05, lim_max + r * 0.05]
    else:
        # only extend upper bound
        return [0, lim_max + r * 0.05]




def frame_lim(mod, func, frames):
    '''
    Find a large enough xlim and ylim for frames, if not heatmaps.

    Inputs:
        mod:        A simulation.model object, the simulation results.
        frames:     How many frame to make for the video.
    
    Returns:
        xlim and ylim for U and V, 4 in total.
    '''
    
    # take 10 samples and store their lims in list
    U_xlist = []
    U_ylist = []
    V_xlist = []
    V_ylist = []
    
    for i in range(10):
        fig_U, ax_U = plt.subplots()
        fig_V, ax_V = plt.subplots()
        ax_U, ax_V = func(mod, ax_U = ax_U, ax_V = ax_V, start = i / 10, end = (i / 10 + 1 / frames))

        U_xlist.append(ax_U.get_xlim())
        U_ylist.append(ax_U.get_ylim())
        V_xlist.append(ax_V.get_xlim())
        V_ylist.append(ax_V.get_ylim())

        plt.close(fig_U)
        plt.close(fig_V)
    
    # get the largest 'range' based on the lists
    U_xlim = get_max_lim(U_xlist)
    U_ylim = get_max_lim(U_ylist)
    V_xlim = get_max_lim(V_xlist)
    V_ylim = get_max_lim(V_ylist)
    
    return U_xlim, U_ylim, V_xlim, V_ylim




def frame_heatmap_lim(mod, func, frames):
    '''
    Find a large enough color bar lim for frames, if heatmaps.

    Inputs:
        mod:        A simulation.model object, the simulation results.
        frames:     How many frame to make for the video.
    
    Returns:
        clim for U and V
    '''

    U_list = []
    V_list = []

    for i in range(10):
        fig_U, ax_U = plt.subplots()
        fig_V, ax_V = plt.subplots()
        func(mod, ax_U = ax_U, ax_V = ax_V, start = i / 10, end = (i / 10 + 1 / frames))

        U_list.append(ax_U.images[0].get_clim())
        V_list.append(ax_V.images[0].get_clim())

        plt.close(fig_U)
        plt.close(fig_V)

    U_clim = get_max_lim(U_list)
    V_clim = get_max_lim(V_list)

    return U_clim, V_clim




def make_mp4(video_dir, frame_dir, fps):
    '''
    Read .png from the frames folder and make into a mp4
    Inputs:
        video_dir:  where to save the video
        frame_dirs: where to read frames from
        fps:        frames per second
    '''

    frame_paths_incomplete = os.listdir(frame_dir)
    frame_paths_incomplete.sort(key=lambda f: int(''.join(filter(str.isdigit, f))))
    frame_path = []
    for file in frame_paths_incomplete:
        if (file[-4:] == '.png') and ('frame' in file):
            frame_path.append(os.path.join(frame_dir, file))

    first_frame = imread(frame_path[0])
    height, width, _ = first_frame.shape
    fourcc = VideoWriter_fourcc(*'mp4v')
    video_writer = VideoWriter(video_dir, fourcc, fps, (width, height))

    for file in frame_path:
            frame = imread(file)
            video_writer.write(frame)
    video_writer.release()




def make_video(mod, func_name = 'UV_hmap', frames = 100, dpi = 200, fps = 30, U_color = 'Greens', V_color = 'Purples', del_frames = False, dirs = 'videos'):
    '''
    Make a mp4 video based on simulation results.

    Inputs:
    - mod:            a simulation.model object, the simulation results.
    - func_name:      what function to use to make the frames. Should be one of the functions in figures.py
    - frames:         how many frames to make. Use more frames for more smooth evolutions.
    - dpi:            dots per inch.
    - fps:            frames per second.
    - U_color:        color for U's videos. Color maps or regular colors, based on what function you use.
    - V_color:        color for V's videos.
    - del_frames:     whether to delete frames after making video.
    - dirs:           where to store the frames and videos.
    '''
    
    if func_name not in FUNC_DICT.keys():
        raise ValueError(func_name + ' not supported for videos.')
    func = FUNC_DICT[func_name]

    # convert color if invalid colors are given
    U_color, V_color = convert_color(func_name, U_color, V_color)

    # set fast style for faster speed
    original_rc = mpl.rcParams.copy()
    mplstyle.use('fast')
    
    # print progress
    one_progress = frames / 100
    current_progress = one_progress
    
    if 'hmap' in func_name:
        # make sure a fixed color bar for all frames
        U_clim, V_clim = frame_heatmap_lim(mod, func, frames)
    else:
        # make sure y axis not changing if not making heatmaps
        U_xlim, U_ylim, V_xlim, V_ylim = frame_lim(mod, func, frames)

    
    U_frame_dirs = dirs + '/U-' + func_name
    V_frame_dirs = dirs + '/V-' + func_name
    
    if os.path.exists(U_frame_dirs):
        file_t.del_dirs(U_frame_dirs)
    os.makedirs(U_frame_dirs)
    if os.path.exists(V_frame_dirs):
        file_t.del_dirs(V_frame_dirs)
    os.makedirs(V_frame_dirs)

    figsize = (6.4, 4.8)
        

    #### for loop ####
    
    for i in range(frames - 1):
        if i > current_progress:
            print('making frames', round(i / frames * 100), '%', end = '\r')
            current_progress += one_progress
        
        if ('bar' in func_name) and (mod.M > 60):
            figsize = (min(mod.M * 0.12, 7.2), 4.8)
            fig_U, ax_U = plt.subplots(figsize = figsize)
            fig_V, ax_V = plt.subplots(figsize = figsize)
        else:
            fig_U, ax_U = plt.subplots(figsize = figsize)#, constrained_layout = True)
            fig_V, ax_V = plt.subplots(figsize = figsize)#, constrained_layout = True)

        if 'hmap' in func_name:
            func(mod, ax_U = ax_U, ax_V = ax_V, U_color = U_color, V_color = V_color, start = i / frames, end = (i + 1) / frames, vrange_U = U_clim, vrange_V = V_clim)
        else:
            func(mod, ax_U = ax_U, ax_V = ax_V, U_color = U_color, V_color = V_color, start = i / frames, end = (i + 1) / frames)
        
        if 'hmap' in func_name:
            # color map lim already set at function call
            pass
        else:
            # make sure y axis not changing if not heatmap and not UV_pi
            ax_U.set_ylim(U_ylim)
            ax_V.set_ylim(V_ylim)
            if ('hist' in func_name) or (func_name == 'UV_pi'):
                # need to set xlim as well for UV_pi and histograms
                ax_V.set_xlim(U_xlim)
                ax_V.set_xlim(V_xlim)

        fig_U.savefig(U_frame_dirs + '/' + 'U_frame_' + str(i) + '.png', pad_inches = 0.25, dpi = dpi)
        fig_V.savefig(V_frame_dirs + '/' + 'V_frame_' + str(i) + '.png', pad_inches = 0.25, dpi = dpi)
        
        plt.close(fig_U)
        plt.close(fig_V)
        
    #### for loop ends ####
    
    # frames done
    print('making mp4...      ', end = '\r')
    
    # make videos based on frames
    make_mp4(dirs + '/U-' + func_name + '.mp4', U_frame_dirs, fps)
    make_mp4(dirs + '/V-' + func_name + '.mp4', V_frame_dirs, fps)
    
    if del_frames:
        file_t.del_dirs(U_frame_dirs)
        file_t.del_dirs(V_frame_dirs)
        print('video saved: ' + dirs + ', frames deleted')
    else:
        print('video saved: ' + dirs + '      ')






'''
Payoff-Driven Stochastic Spatial Model for Evolutionary Game Theory
-----------------------------------------------------------------------------

Provides:
    1.  A stochastic spatial model for simulating the interaction and evolution of two species in either 1D or 2D space
    2.  Plot & video functions to visualize simulation results.
    3.  Module to test influence of certain variables on results.
    4.  Data saving & reading module.
    4.  Additional analytical tools.

Websites:
    - The *piegy* documentation: https://piegy.readthedocs.io/en/
    - GitHub repository at: https://github.com/Chenning04/piegy.git
    - PyPI page: https://pypi.org/project/piegy/


Last update: May 12, 2025
'''

from .__version__ import __version__
from .build_info import build_info

from .simulation import model, run, demo_model, UV_expected_val, check_overflow_func
from .videos import make_video, SUPPORTED_FIGURES
from .data_tools import save_data, read_data

from .analysis import rounds_expected, scale_maxtime, check_convergence, combine_sim

from .figures import (UV_hmap, UV_bar, UV_dyna, UV_hist, UV_std, UV_expected, 
                      pi_hmap, pi_bar, pi_dyna, pi_hist, pi_std, UV_pi)

from .test_var import (test_var1, var_UV1, var_pi1, var_convergence1, get_dirs1, 
                       test_var2, var_UV2, var_pi2, var_convergence2, get_dirs2)


simulation_memebers = ['model', 'run', 'demo_model']

videos_members = ['make_video', 'SUPPORTED_FIGURES']

data_members = ['save_data', 'read_data']

analysis_members = ['expected_rounds', 'scale_maxtime', 'check_convergence', 'combine_mod']

figures_members = ['UV_hmap', 'UV_bar', 'UV_dyna', 'UV_hist', 'UV_std', 'UV_expected_val', 'UV_expected', 
                   'pi_hmap', 'pi_bar', 'pi_dyna', 'pi_hist', 'pi_std', 'UV_pi']

test_var_members = ['test_var1', 'var_UV1', 'var_pi1', 'var_convergence1', 'get_dirs1', 
                    'test_var2', 'var_UV2', 'var_pi2', 'var_convergence2', 'get_dirs2']


__all__ = simulation_memebers + videos_members + data_members + figures_members + analysis_members + test_var_members








'''
Init file for .tools sub-package.

Contains three tool kits:
    - data_tools:   intended for both direct usage and private helper module.
    - file_tools:   private use.
    - figure_tools: private use.

    
Last updated: Mar 27, 2025
'''

from . import file_tools
from . import figure_tools




"""
Contains build info, whether it's local built, or a pre-compiled wheel.
Auto-generated at compile time.
"""

build_info = {
    "version": "version",
    "build date": "build data",
    "python used": "python version",
    "platform": "platform"
}



'''
Functions and class below are for Python-based simulations. Not maintained after v1.1.6 on Jun 26, 2025.
But you can still run them by calling:

>>> run_py(mod)
'''

from .simulation import model

import math
import numpy as np
from timeit import default_timer as timer


class patch:
    '''
    A single patch in the N x M space.
    Interacts with neighboring patches, assuming no spatial structure within a patch.
    Initialized in single_init function.

    Class Functions:

        __init__:
            Inputs:
                U, V:       initial value of U and V
                matrix:     payoff matrix for U and V. The canonical form is 2x2, here we ask for a flattened 1x4 form.
                patch_var:  np.array of [mu1, mu2, w1, w2, kappa1, kappa2]
        
        __str__:
            Print patch object in a nice way.

        set_nb_pointers:
            Set pointers to neighbors of this patch object.

        update_pi:
            Update Upi, Vpi and payoff rates (payoff rates are the first two numbers in self.pi_death_rates).
        
        update_k:
            Update natural death rates (the last two numbers in self.pi_death_rates).

        update_mig:
            Update migration rates.

        get_pi_death_rates, get_mig_rates:
            Return respective members.

        change_popu:
            Change U, V based on input signal.
    '''
    
    def __init__(self, U, V, matrix = [-0.1, 0.4, 0, 0.2], patch_var = [0.5, 0.5, 100, 100, 0.001, 0.001]):

        self.U = U                          # int, U population. Initialized upon creating object.
        self.V = V                          # int, V population
        self.Upi = 0                       # float, payoff
        self.Vpi = 0

        self.matrix = matrix                # np.array or list, len = 4, payoff matrix
        self.mu1 = patch_var[0]             # float, how much proportion of the population migrates (U) each time
        self.mu2 = patch_var[1]
        self.w1 = patch_var[2]              # float, strength of payoff-driven effect. Larger w <=> stronger payoff-driven motion
        self.w2 = patch_var[3]
        self.kappa1 = patch_var[4]          # float, carrying capacity, determines death rates
        self.kappa2 = patch_var[5]

        self.nb = None                                  # list of patch objects (pointers), point to neighbors, initialized seperatedly (after all patches are created)
        self.pi_death_rates = [0 for _ in range(4)]     # list, len = 4, rates of payoff & death
                                                        # first two are payoff rates, second two are death rates
        self.mig_rates = [0 for _ in range(8)]          # list, len = 8, migration rates, stored in an order: up, down, left, right, 
                                                        # first 4 are U's mig_rate, the last 4 are V's
        self.sum_pi_death_rates = 0                     # float, sum of pi_death_rates
        self.sum_mig_rates = 0                          # float, sum of mig_rates
        

    def __str__(self):
        self_str = ''
        self_str += 'U, V = ' + str(self.U) + ', ' + str(self.V) + '\n'
        self_str += 'pi = ' + str(self.Upi) + ', ' + str(self.Vpi) + '\n'
        self_str += 'matrix = ' + str(self.matrix) + '\n'
        self_str += 'mu1, mu2 = ' + str(self.mu1) + ', ' + str(self.mu2) + '\n'
        self_str += 'w1, w2 = ' + str(self.w1) + ', ' + str(self.w2) + '\n'
        self_str += 'kappa1, kappa2 = ' + str(self.kappa1) + ', ' + str(self.kappa2) + '\n'
        self_str += '\n'
        self_str += 'nb = ' + str(self.nb)
        self_str += 'pi_death_rates = ' + str(self.pi_death_rates) + '\n'
        self_str += 'mig_rates = ' + str(self.mig_rates) + '\n'
        self_str += 'sum_pi_death_rates = ' + str(self.sum_pi_death_rates) + '\n'
        self_str += 'sum_mig_rates = ' + str(self.sum_mig_rates) + '\n'

        return self_str
    
    
    def set_nb_pointers(self, nb):
        # nb is a list of pointers (point to patches)
        # nb is passed from the model class
        self.nb = nb
    
    
    def update_pi_k(self):
        # calculate payoff and natural death rates

        U = self.U  # bring the values to front
        V = self.V
        sum_minus_1 = U + V - 1  # this value is used several times
        
        if sum_minus_1 > 0:
            # interaction happens only if there is more than 1 individual
            
            if U != 0:
                # no payoff if U == 0
                self.Upi = (U - 1) / sum_minus_1 * self.matrix[0] + V / sum_minus_1 * self.matrix[1]
            else:
                self.Upi = 0
                
            if V != 0:
                self.Vpi = U / sum_minus_1 * self.matrix[2] + (V - 1) / sum_minus_1 * self.matrix[3]
            else:
                self.Vpi = 0
                
        else:
            # no interaction, hence no payoff, if only 1 individual
            self.Upi = 0
            self.Vpi = 0

        # update payoff rates
        self.pi_death_rates[0] = abs(U * self.Upi)
        self.pi_death_rates[1] = abs(V * self.Vpi)

        # update natural death rates
        self.pi_death_rates[2] = self.kappa1 * U * (sum_minus_1 + 1)
        self.pi_death_rates[3] = self.kappa2 * V * (sum_minus_1 + 1)

        # update sum of rates
        self.sum_pi_death_rates = sum(self.pi_death_rates)
    

    def update_mig(self):
        # calculate migration rates

        # store the 'weight' of migration, i.e. value of f/g functions for neighbors
        U_weight = [0, 0, 0, 0]
        V_weight = [0, 0, 0, 0]

        for i in range(4):
            if self.nb[i] != None:
                U_weight[i] = 1 + pow(math.e, self.w1 * self.nb[i].Upi)
                V_weight[i] = 1 + pow(math.e, self.w2 * self.nb[i].Vpi)

        mu1_U = self.mu1 * self.U
        mu2_V = self.mu2 * self.V

        mu1_U_divide_sum = mu1_U / sum(U_weight)
        mu2_V_divide_sum = mu2_V / sum(V_weight)
        
        for i in range(4):
            self.mig_rates[i] = mu1_U_divide_sum * U_weight[i]
            self.mig_rates[i + 4] = mu2_V_divide_sum * V_weight[i]

        # update sum of rates
        self.sum_mig_rates = mu1_U + mu2_V
    

    def get_sum_rates(self):
        # return sum of all 12 rates
        return self.sum_pi_death_rates + self.sum_mig_rates


    def find_event(self, expected_sum):
        # find the event within the 12 events based on expected sum-of-rates within this patch

        if expected_sum > (self.sum_pi_death_rates + self.sum_mig_rates):
            print("patch rate not large enough")

        if expected_sum < self.sum_pi_death_rates:
            # in the first 4 events (payoff and death events)
            event = 0
            current_sum = 0
            while current_sum < expected_sum:
                current_sum += self.pi_death_rates[event]
                event += 1
            event -= 1

        else:
            # in the last 8 events (migration events):
            event = 0
            current_sum = self.sum_pi_death_rates
            while current_sum < expected_sum:
                current_sum += self.mig_rates[event]
                event += 1
            event += 3  # i.e., -= 1, then += 4 (to account for the first 4 payoff & death rates)

        return event

    
    def change_popu(self, s):
        # convert s (a signal, passed from model class) to a change in population
        
        # s = 0, 1, 2 are for U
        # s = 0 for migration IN, receive an immigrant
        if s == 0:
            self.U += 1       # receive an immigrant
        # s = 1 for migration OUT / death due to carrying capacity
        elif s == 1:
            if self.U > 0:
                self.U -= 1
        # s = 2 for natural birth / death, due to payoff
        elif s == 2:
            if self.Upi > 0:
                self.U += 1   # natural growth due to payoff
            elif self.U > 0:
                self.U -= 1   # natural death due to payoff
        
        # s = 3, 4, 5 are for V
        elif s == 3:
            self.V += 1
        elif s == 4:
            if self.V > 0:
                self.V -= 1
        else:
            if self.Vpi > 0:
                self.V += 1
            elif self.V > 0:
                self.V -= 1



def find_nb_zero_flux(N, M, i, j):
    '''
    Find neighbors of patch (i, j) in zero-flux boundary condition. i.e., the space is square with boundary.
    Return neighbors' indices in an order: up, down, left, right.
    Index will be None if no neighbor exists in that direction.
    '''
    nb_indices = []
    
    if i != 0: 
        nb_indices.append([i - 1, j])  # up
    else:
        nb_indices.append(None)    # neighbor doesn't exist
    
    if i != N - 1:
        nb_indices.append([i + 1, j])  # down
    else:
        nb_indices.append(None)
    
    if j != 0:
        nb_indices.append([i, j - 1])  # left
    else:
        nb_indices.append(None)
    
    if j != M - 1:
        nb_indices.append([i, j + 1])  # right
    else:
        nb_indices.append(None)
    
    return nb_indices




def find_nb_periodical(N, M, i, j):
    '''
    Find neighbors of patch (i, j) in periodical boundary condition. i.e., the space is a sphere.
    Return neighbors' indices in an order: up, down, left, right.
    If space not 1D, a neighbor always exists.
    If space is 1D, say N = 1, we don't allow (0, j) to migrate up & down (self-self migration is considered invalid)
    '''
    nb_indices = []
    
    # up
    if N != 1:
        if i != 0: 
            nb_indices.append([i - 1, j])
        else:
            nb_indices.append([N - 1, j])
    else:
        nb_indices.append(None)  # can't migrate to itself
    
    # down
    if N != 1:
        if i != N - 1:
            nb_indices.append([i + 1, j])
        else:
            nb_indices.append([0, j])
    else:
        nb_indices.append(None)
    
    # left
    # No need to check M == 1 because we explicitly asked for M > 1
    if j != 0:
        nb_indices.append([i, j - 1])
    else:
        nb_indices.append([i, M - 1])
    
    # right
    if j != M - 1:
        nb_indices.append([i, j + 1])
    else:
        nb_indices.append([i, 0])
        
    return nb_indices




def find_patch(expected_sum, patch_rates, sum_rates_by_row, sum_rates):
    '''
    Find which patch the event is in. Only patch index is found, patch.find_event find which event it is exactly.

    Inputs:
        expected_sum:       a random number * sum of all rates. Essentially points to a random event.
                            We want to find the patch that contains this pointer.
        patch_rates:        a N x M np.array. Stores sum of the 12 rates in every patch.
        sum_rates_by_row:   a 1D np.array with len = N. Stores the sum of the M x 12 rates in every row.
        sum_rates:          sum of all N x M x 12 rates.
    
    Returns:
        row, col:           row and column number of where the patch.
    '''
    
    # Find row first
    if expected_sum < sum_rates / 2:
        # search row forwards if in the first half of rows
        current_sum = 0
        row = 0
        while current_sum < expected_sum:
            current_sum += sum_rates_by_row[row]
            row += 1
        row -= 1
        current_sum -= sum_rates_by_row[row]  # need to subtract that row (which caused current sum to exceed expected_sum)
    else:
        # search row backwards if in the second half of rows
        current_sum = sum_rates
        row = len(patch_rates) - 1
        while current_sum > expected_sum:
            current_sum -= sum_rates_by_row[row]
            row -= 1
        row += 1
        # don't need subtraction here, as current_sum is already < expected same

    # Find col in that row
    if (expected_sum - current_sum) < sum_rates_by_row[row] / 2:
        # search col forwards if in the first half of that row
        col = 0
        while current_sum < expected_sum:
            current_sum += patch_rates[row][col]
            col += 1
        col -= 1
        current_sum -= patch_rates[row][col]  # need a subtraction
    else:
        # search col backwards if in the second half of that row
        current_sum += sum_rates_by_row[row]
        col = len(patch_rates[0]) - 1
        while current_sum > expected_sum:
            current_sum -= patch_rates[row][col]
            col -= 1
        col += 1
        # don't need subtraction

    return row, col, current_sum




def make_signal_zero_flux(i, j, e):
    '''
    Find which patch to change what based on i, j, e (event number) value, for the zero-flux boundary condition

    Inputs: 
        i, j is the position of the 'center' patch, e is which event to happen there.
        Another patch might be influenced as well if a migration event was picked.

        Possible values for e:
            e = 0 or 1: natural change of U/V due to payoff. 
                        Can be either brith or death (based on payoff is positive or negative).
                        Cooresponds to s = 2 or 5 in the patch class
            e = 2 or 3: death of U/V due to carrying capacity. 
                        Cooresponds to s = 1 or 4 in patch: make U/V -= 1
            e = 4 ~ 7:  migration events of U, patch (i, j) loses an individual, and another patch receives one.
                        we use the up-down-left-right rule for the direction. 4 means up, 5 means down, ...
                        Cooresponds to s = 0 for the mig-in patch (force U += 1), and s = 1 for the mig-out patch (force U -= 1)
            e = 8 ~ 11: migration events of V.
                        Cooresponds to s = 3 for the mig-in patch (force V += 1), and s = 4 for the mig-out patch (force V -= 1)
    '''
    if e < 6:
        if e == 0:
            return [[i, j, 2]]
        elif e == 1:
            return [[i, j, 5]]
        elif e == 2:
            return [[i, j, 1]]
        elif e == 3:
            return [[i, j, 4]]
        elif e == 4:
            return [[i, j, 1], [i - 1, j, 0]]
        else:
            return [[i, j, 1], [i + 1, j, 0]]
    else:
        if e == 6:
            return [[i, j, 1], [i, j - 1, 0]]
        elif e == 7:
            return [[i, j, 1], [i, j + 1, 0]]
        elif e == 8:
            return [[i, j, 4], [i - 1, j, 3]]
        elif e == 9:
            return [[i, j, 4], [i + 1, j, 3]]
        elif e == 10:
            return [[i, j, 4], [i, j - 1, 3]]
        elif e == 11:
            return [[i, j, 4], [i, j + 1, 3]]
        else:
            raise RuntimeError('A bug in code: invalid event number encountered:', e)  # debug line



def make_signal_periodical(N, M, i, j, e):
    '''
    Find which patch to change what based on i, j, e value, for the periodical boundary condition
    Similar to make_signal_zero_flux.
    '''
    
    if e < 6:
        if e == 0:
            return [[i, j, 2]]
        elif e == 1:
            return [[i, j, 5]]
        elif e == 2:
            return [[i, j, 1]]
        elif e == 3:
            return [[i, j, 4]]
        elif e == 4:
            if i != 0:
                return [[i, j, 1], [i - 1, j, 0]]
            else:
                return [[i, j, 1], [N - 1, j, 0]]
        else:
            if i != N - 1:
                return [[i, j, 1], [i + 1, j, 0]]
            else:
                return [[i, j, 1], [0, j, 0]]
    else:
        if e == 6:
            if j != 0:
                return [[i, j, 1], [i, j - 1, 0]]
            else:
                return [[i, j, 1], [i, M - 1, 0]]
        elif e == 7:
            if j != M - 1:
                return [[i, j, 1], [i, j + 1, 0]]
            else:
                return [[i, j, 1], [i, 0, 0]]
        elif e == 8:
            if i != 0:
                return [[i, j, 4], [i - 1, j, 3]]
            else:
                return [[i, j, 4], [N - 1, j, 3]]
        elif e == 9:
            if i != N - 1:
                return [[i, j, 4], [i + 1, j, 3]]
            else:
                return [[i, j, 4], [0, j, 3]]
        elif e == 10:
            if j != 0:
                return [[i, j, 4], [i, j - 1, 3]]
            else:
                return [[i, j, 4], [i, M - 1, 3]]
        elif e == 11:
            if j != M - 1:
                return [[i, j, 4], [i, j + 1, 3]]
            else:
                return [[i, j, 4], [i, 0, 3]]
        else:
            raise RuntimeError('A bug in code: invalid event number encountered:', e)  # debug line




def nb_need_change(ni, signal):
    '''
    Check whether a neighbor needs to change.
    Two cases don't need change: either ni is None (doesn't exist) or in signal (is a last-change patch and already updated)

    Inputs:
        ni:         index of a neighbor, might be None if patch doesn't exist.
        signal:     return value of make_signal_zero_flux or make_signal_periodical.
    
    Returns:
        True or False, whether the neighboring patch specified by ni needs change
    '''
    
    if ni == None:
        return False
    
    for si in signal:
        if ni[0] == si[0] and ni[1] == si[1]:
            return False
    
    return True




def single_init(mod, rng):
    '''
    The first major function for the model.
    Initialize all variables and run 1 round, then pass variables and results to single_test.

    Input:
        mod is a model object
        rng is random number generator (np.random.default_rng), initialized by model.run
    '''

    #### Initialize Data Storage ####

    world = [[patch(mod.I[i][j][0], mod.I[i][j][1], mod.X[i][j], mod.P[i][j]) for j in range(mod.M)] for i in range(mod.N)]  # N x M patches
    patch_rates = np.zeros((mod.N, mod.M), dtype = np.float64)  # every patch's sum-of-12-srates
    sum_rates_by_row = np.zeros((mod.N), dtype = np.float64)  # every row's sum-of-patch, i.e., sum of 12 * M rates in every row.
    sum_rates = 0  # sum of all N x M x 12 rates

    signal = None

    nb_indices = None
    if mod.boundary:
        nb_indices = [[find_nb_zero_flux(mod.N, mod.M, i, j) for j in range(mod.M)] for i in range(mod.N)]
    else:
        nb_indices = [[find_nb_periodical(mod.N, mod.M, i, j) for j in range(mod.M)] for i in range(mod.N)]
        
    for i in range(mod.N):
        for j in range(mod.M):
            nb = []
            for k in range(4):
                if nb_indices[i][j][k] != None:
                    # append a pointer to the patch
                    nb.append(world[nb_indices[i][j][k][0]][nb_indices[i][j][k][1]])
                else:
                    # nb doesn't exist
                    nb.append(None)
            # pass it to patch class and store
            world[i][j].set_nb_pointers(nb)
    

    #### Begin Running ####

    # initialize payoff & natural death rates
    for i in range(mod.N):
        for j in range(mod.M):
            world[i][j].update_pi_k()
    
    # initialize migration rates & the rates list
    for i in range(mod.N):
        for j in range(mod.M):
            world[i][j].update_mig()
            # store rates & sum of rates
            patch_rates[i][j] = world[i][j].get_sum_rates()
        sum_rates_by_row[i] = sum(patch_rates[i])

    sum_rates = sum(sum_rates_by_row)
    
    # pick the first random event
    expected_sum = rng.random() * sum_rates
    # find patch first
    i0, j0, current_sum = find_patch(expected_sum, patch_rates, sum_rates_by_row, sum_rates)
    # then find which event in that patch
    e0 = world[i0][j0].find_event(expected_sum - current_sum)

    # initialize signal
    if mod.boundary:
        signal = make_signal_zero_flux(i0, j0, e0)   # walls around world
    else:
        signal = make_signal_periodical(mod.N, mod.M, i0, j0, e0)  # no walls around world
    
    # change U&V based on signal
    for si in signal:
        world[si[0]][si[1]].change_popu(si[2])
        
    # time increment
    time = (1 / sum_rates) * math.log(1 / rng.random())
    
    # record
    if time > mod.record_itv:
        record_index = int(time / mod.record_itv)
        for i in range(mod.N):
            for j in range(mod.M):
                for k in range(record_index):
                    mod.U[i][j][k] += world[i][j].U
                    mod.V[i][j][k] += world[i][j].V
                    mod.Upi[i][j][k] += world[i][j].Upi
                    mod.Vpi[i][j][k] += world[i][j].Vpi
                    # we simply add to that entry, and later divide by sim_time to get the average (division in run function)
    
    return time, world, nb_indices, patch_rates, sum_rates_by_row, sum_rates, signal




def single_test(mod, front_info, end_info, update_sum_frequency, rng):
    '''
    Runs a single model, from time = 0 to mod.maxtime.
    run recursively calls single_test to get the average data. 

    Inputs:
        sim:                    a model object, created by user and carries all parameters & storage bins.
        front_info, end_info:   passed by run to show messages, like the current round number in run. Not intended for direct usages.
        update_sum_frequency:   re-calculate sums this many times in model. 
                                Our sums are gradually updated over time. So might have precision errors for large maxtime.
        rng:                    np.random.default_rng. Initialized by model.run
    '''

    # initialize helper variables
    # used to print progress, i.e., how much percent is done
    one_time = mod.maxtime / max(100, update_sum_frequency) 
    one_progress = 0
    if mod.print_pct != None:
        # print progress, x%
        print(front_info +  ' 0%' + end_info, end = '\r')
        one_progress = mod.maxtime * mod.print_pct / 100
    else:
        one_progress = 2 * mod.maxtime     # not printing

    # our sums (sum_rates_by_row and sum_rates) are gradually updated over time. This may have precision errors for large maxtime.
    # So re-sum everything every some percentage of maxtime.
    one_update_sum = mod.maxtime / update_sum_frequency

    current_time = one_time
    current_progress = one_progress
    current_update_sum = one_update_sum

    max_record = int(mod.maxtime / mod.record_itv)

    
    # initialize
    time, world, nb_indices, patch_rates, sum_rates_by_row, sum_rates, signal = single_init(mod, rng)
    record_index = int(time / mod.record_itv)
    # record_time is how much time has passed since the last record
    # if record_time > record_itv:
    #    we count how many record_itvs are there in record_time, denote the number by multi_records
    #    then store the current data in multi_records number of cells in the list
    #    and subtract record_time by the multiple of record_itv, so that record_time < record_itv
    record_time = time - record_index * mod.record_itv

    ### Large while loop ###
    
    while time < mod.maxtime:
    
        # print progress & correct error of sum_rates
        if time > current_time:
            # a new 1% of time
            current_time += one_time
            if time > current_progress:
                # print progress
                print(front_info +  ' ' + str(round(time / mod.maxtime * 100)) + '%' + end_info, end = '\r')
                current_progress += one_progress

            if time > current_update_sum:
                current_update_sum += one_update_sum
                for i in range(mod.N):
                    sum_rates_by_row[i] = sum(patch_rates[i])
                sum_rates = sum(sum_rates_by_row)

        
        # before updating last-changed patches, subtract old sum of rates (so as to update sum of rates by adding new rates later)
        for si in signal:
            # si[0] is row number, si[1] is col number
            old_patch_rate = world[si[0]][si[1]].get_sum_rates()
            sum_rates_by_row[si[0]] -= old_patch_rate
            sum_rates -= old_patch_rate

        # update last-changed patches
        # update payoff and death rates first
        for si in signal:
            world[si[0]][si[1]].update_pi_k()
        # then update migration rates, as mig_rates depend on neighbor's payoff
        for si in signal:
            world[si[0]][si[1]].update_mig()
            
            # update rates stored
            new_patch_rate = world[si[0]][si[1]].get_sum_rates()
            # update patch_rates
            patch_rates[si[0]][si[1]] = new_patch_rate
            # update sum_rate_by_row and sum_rates_by_row by adding new rates
            sum_rates_by_row[si[0]] += new_patch_rate
            sum_rates += new_patch_rate

        # update neighbors of last-changed patches
        for si in signal:
            for ni in nb_indices[si[0]][si[1]]:
                # don't need to update if the patch is a last-change patch itself or None
                # use helper function to check
                if nb_need_change(ni, signal):
                    # update migratino rates
                    world[ni[0]][ni[1]].update_mig()
                    # Note: no need to update patch_rates and sum of rates, as update_mig doesn't change total rates in a patch.
                    # sum_mig_rate is decided by mu1 * U + mu2 * V, and pi_death_rate is not changed.
               
        # pick the first random event
        expected_sum = rng.random() * sum_rates
        # find patch first
        i0, j0, current_sum = find_patch(expected_sum, patch_rates, sum_rates_by_row, sum_rates)
        # then find which event in that patch
        e0 = world[i0][j0].find_event(expected_sum - current_sum)

        # make signal
        if mod.boundary:
            signal = make_signal_zero_flux(i0, j0, e0)
        else:
            signal = make_signal_periodical(mod.N, mod.M, i0, j0, e0)

        # let the event happen
        for si in signal:
            world[si[0]][si[1]].change_popu(si[2])
        
        # increase time
        r1 = rng.random()
        dt = (1 / sum_rates) * math.log(1 / r1)
        time += dt
        record_time += dt
    
        if time < mod.maxtime:
            # if not exceeds maxtime
            if record_time > mod.record_itv:
                multi_records = int(record_time / mod.record_itv)
                record_time -= multi_records * mod.record_itv

                for i in range(mod.N):
                    for j in range(mod.M):
                        for k in range(record_index, record_index + multi_records):
                            mod.U[i][j][k] += world[i][j].U
                            mod.V[i][j][k] += world[i][j].V
                            mod.Upi[i][j][k] += world[i][j].Upi
                            mod.Vpi[i][j][k] += world[i][j].Vpi
                record_index += multi_records
        else:
            # if already exceeds maxtime
            for i in range(mod.N):
                for j in range(mod.M):
                    for k in range(record_index, max_record):
                        mod.U[i][j][k] += world[i][j].U
                        mod.V[i][j][k] += world[i][j].V
                        mod.Upi[i][j][k] += world[i][j].Upi
                        mod.Vpi[i][j][k] += world[i][j].Vpi

    ### Large while loop ends ###
    
    if mod.print_pct != None:
        print(front_info + ' 100%' + ' ' * 20, end = '\r')  # empty spaces to overwrite predicted runtime




def run_py(mod, predict_runtime = False, message = ''):
    '''
    Main function. Recursively calls single_test to run many models and then takes the average.

    Inputs:
    - mod is a model object.
    - predict_runtime = False will not predict how much time still needed, set to True if you want to see.
    - message is used by some functions in figures.py to print messages.
    '''

    if not mod.data_empty:
        raise RuntimeError('mod has non-empty data')
    
    mod.U = np.zeros((mod.N, mod.M, mod.max_record))
    mod.V = np.zeros((mod.N, mod.M, mod.max_record))
    mod.Vpi = np.zeros((mod.N, mod.M, mod.max_record))
    mod.Upi = np.zeros((mod.N, mod.M, mod.max_record))
    
    start = timer()   # runtime

    mod.data_empty = False
    rng = np.random.default_rng(mod.seed)
    
    # passed to single_test to print progress
    if mod.print_pct == 0:
        mod.print_pct = 5    # default print_pct

    update_sum_frequency = 4   # re-calculate sums this many times. See input desciption of single_test
    
    ### models ###
    i = 0
    
    while i < mod.sim_time:
        # use while loop so that can go backwards if got numerical issues

        end_info = ''
        if predict_runtime:
            if i > 0:
                time_elapsed = timer() - start
                pred_runtime = time_elapsed / i * (mod.sim_time - i)
                end_info = ', ~' + str(round(pred_runtime, 2)) + 's left'
        
        front_info = ''
        if mod.print_pct != None:
            front_info = message + 'round ' + str(i) + ':'
            print(front_info + ' ' * 30, end = '\r')  # the blank spaces are to overwrite percentages, e.g. 36 %
        
        try:
            single_test(mod, front_info, end_info, update_sum_frequency, rng)
            i += 1
        except IndexError:
            update_sum_frequency *= 4
            print('Numerical issue at round ' + str(i) + '. Trying higher precision now. See doc if err repeats')
            # not increasing i: redo current round.
        
    ### models end ###

    mod.calculate_ave()
    
    stop = timer()
    print(' ' * 30, end = '\r')  # overwrite all previous prints
    print(message + 'runtime: ' + str(round(stop - start, 2)) + ' s')
    return








'''
File-processing tools.

Functions:
- del_dirs:     Delete everything in a directory, as well as the directory itself.
'''

import os


def del_dirs(dirs):
    # Delete everything in a directory.
    
    subdirs_list = []
    
    for subdirs, dirs_, files in os.walk(dirs):
        if subdirs not in subdirs_list:
            subdirs_list.append(subdirs)
            
        for file in files:
            path = os.path.join(subdirs, file)
            if os.path.isfile(path):
                os.remove(path)
    
    len_s = len(subdirs_list)
    
    for i in range(len_s):
        os.rmdir(subdirs_list[len_s - i - 1])




'''
This module is used to find the C core
'''

import os


def find_C():
    C_core_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'C_core')
    C_list = os.listdir(C_core_path)
    for file in C_list:
        if file[-3:] == '.so':
            return C_core_path + '/' + file
        if file[-4:] == '.pyd':
            return C_core_path + '/' + file

    raise FileNotFoundError('C computation core not found. You can either compile manully or use the Python core instead. Please see docs.')



import subprocess
import sys
import os
import re
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
import platform
import datetime
import textwrap


class BuildSharedLib(build_ext):
    def run(self):
        # Build your shared lib first
        subprocess.check_call(['make', '-C', 'src/piegy/C_core', 'clean', 'so'])
        # Then continue with normal build_ext run
        super().run()

    def build_extension(self, ext):
        if sys.platform == 'win32':
            lib_name = 'piegyc.pyd'
        else:
            lib_name = 'piegyc.so'
        so_path = os.path.abspath(f'src/piegy/C_core/{lib_name}')
        target_path = self.get_ext_fullpath(ext.name)

        os.makedirs(os.path.dirname(target_path), exist_ok=True)
        self.copy_file(so_path, target_path)
        self.rm_duplicate()

    def rm_duplicate(self):
        C_core_path = os.path.abspath(f'src/piegy/C_core')
        so_name = 'piegyc.so'
        if sys.platform == 'win32':
            so_name = 'piegyc.pyd'
        try:
            os.remove(os.path.join(C_core_path, so_name))
        except OSError:
            pass



class AddCompileInfo(build_py):

    def get_version(self):
        version_file = os.path.join('src', 'piegy', '__version__.py')
        with open(version_file, 'r', encoding='utf-8') as f:
            content = f.read()
        # Regex to extract __version__ = '...'
        version_match = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', content, re.MULTILINE)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")

    def generate_build_info(self):
        version = self.get_version()
        build_info_content = f'''
        """
        Contains build info, whether it's local built, or a pre-compiled wheel.
        Auto-generated at compile time.
        """

        build_info = {{
            "version": "{version}",
            "build date": "{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}",
            "python used": "{platform.python_version()}",
            "platform": "{sys.platform}"
        }}
        '''
        return textwrap.dedent(build_info_content[1:])
    
    
    def run(self):
        # Write build_info.py directly before building python files
        build_info_path = os.path.abspath(os.path.join('src/piegy/build_info.py'))
        with open(build_info_path, 'w') as f:
            f.write(self.generate_build_info())
        super().run()


# Declare the extension module (empty sources)
ext_modules = [
    Extension('piegy.C_core.piegyc', sources=[])
]

setup(
    ext_modules=ext_modules,
    cmdclass={'build_ext': BuildSharedLib, 
              'build_py': AddCompileInfo,},
    package_data={"piegy": ["build_info.py"]},
    include_package_data=True,
)




'''
Main Module of Stochastic Simulation
-------------------------------------------

Contains all the necessary tools to build a model and run models based on Gillespie Algorithm.

Class & Functions:
- model:                creates a stochastic mode. Run the model by ``simulation.run(mod)``
- run:                  runs stochastic simulation on a model.
- demo_model:           returns a demo model.
- UV_expected_val:      calculates the expected population of U, V at steady state, assuming no migration and any stochastic process.
- check_overflow_func:  check whether an overflow might happen in simulation. This is usually done automatically when init-ing a model.
'''

from .tools.find_C import find_C

import numpy as np
import os
import ctypes
from ctypes import c_size_t, c_ubyte, c_uint32, c_int32, c_int64, c_double, c_bool, c_char_p, c_char
import numpy as np
from numpy.ctypeslib import ndpointer


# check whether overflow / too large values might be encountered
# these values are considered as exponents in exp()
EXP_OVERFLOW_BOUND = 709  # where exp(x) reaches overflow bound
EXP_TOO_LARGE_BOUND = 88  # where exp(x) reaches 1e20

# how frequent to update sum in simulation, i.e., accuracy
# the defuault value is 64, increase to 4 times if encounter error 
UPDATE_SUM_FREQ = 64


# read the C core into LIB
# initialized upon first run
LIB = None
# possible result values returned from C
SIM_SUCCESS = 0
SIM_DATA_EMPTY = 1
SIM_SMALL_MAXTIME = 2
SIM_OVERFLOW = 3


'''
The C core
'''

class model_c(ctypes.Structure):
    '''
    The C-cored model
    '''

    _fields_ = [
        ('N', c_size_t),
        ('M', c_size_t),
        ('maxtime', c_double),
        ('record_itv', c_double),
        ('sim_time', c_size_t),
        ('boundary', c_bool),

        ('I', ctypes.POINTER(c_uint32)),
        ('X', ctypes.POINTER(c_double)),
        ('P', ctypes.POINTER(c_double)),

        ('print_pct', c_int32),
        ('seed', c_int32),

        ('data_empty', c_bool),
        ('max_record', c_size_t),
        ('arr_size', c_size_t),
        ('compress_itv', c_uint32),

        ('U1d', ctypes.POINTER(c_double)),
        ('V1d', ctypes.POINTER(c_double)),
        ('Upi_1d', ctypes.POINTER(c_double)),
        ('Vpi_1d', ctypes.POINTER(c_double)),
    ]
    def get_array(self, name):
        """Return internal data as NumPy array, e.g. .get_array('U')"""
        ptr = getattr(self, name)
        return np.ctypeslib.as_array(ptr, shape=(self.arr_size,))


def read_lib():
    global LIB
    if LIB != None:
        return

    LIB = ctypes.CDLL(find_C(), winmode = 0)
    LIB.mod_init.argtypes = [
        ctypes.POINTER(model_c), c_size_t, c_size_t,
        c_double, c_double, c_size_t, c_bool,
        ndpointer(dtype=np.uint32, flags="C_CONTIGUOUS"),
        ndpointer(dtype=np.float64, flags="C_CONTIGUOUS"),
        ndpointer(dtype=np.float64, flags="C_CONTIGUOUS"),
        c_int32, c_int32
    ]
    LIB.mod_init.restype = c_bool

    LIB.mod_free_py.argtypes = [ctypes.POINTER(model_c)]
    LIB.mod_free_py.restype = None

    LIB.run.argtypes = [ctypes.POINTER(model_c), ctypes.POINTER(c_char), c_size_t, c_uint32]
    LIB.run.restype = c_ubyte  # which is uint8





'''
For access by Python
'''             

class model:
    '''
    Store model data and input parameters.
    Initialize a model object to run models.

    Public Class Functions:

        __init__:
            Create a model object. Also initialize data storage.

        __str__:
            Print model object in a nice way.

        copy:
            Return a deep copy of self. Can choose whether to copy data as well. Default is to copy.

        clear_data:
            clear all data stored, set U, V, Upi, Vpi to zero arrays

        change_maxtime:
            Changes maxtime of self. Update data storage as well.

        set_seed:
            Set a new seed. 

        compress_data:
            compress data by only storing average values
    '''

    def __init__(self, N, M, maxtime, record_itv, sim_time, boundary, I, X, P, print_pct = 25, seed = None, check_overflow = True):

        self.check_valid_input(N, M, maxtime, record_itv, sim_time, boundary, I, X, P, print_pct, seed, check_overflow)
        
        self.N = N                      # int, N x M is spatial dimension
        self.M = M                      # int, can't be 1. If want to make 1D space, use N = 1. And this model doesn't work for 1x1 space (causes NaN)
        self.maxtime = maxtime          # float or int, run model for how long time
        self.record_itv = record_itv    # float, record data every record_itv of time
        self.sim_time = sim_time        # int, run this many of rounds (of single_test)
        self.boundary = boundary        # bool, the N x M space have boundary or not (i.e., zero-flux (True) or periodical (False))
        self.I = np.array(I)            # N x M x 2 np.array, initial population. Two init-popu for every patch (U and V)
        self.X = np.array(X)            # N x M x 4 np.array, matrices. The '4' comes from 2x2 matrix flattened to 1D
        self.P = np.array(P)            # N x M x 6 np.array, 'patch variables', i.e., mu1&2, w1&2, kappa1&2
        if print_pct == None:
            self.print_pct = -1
        else:
            self.print_pct = print_pct      # int, print how much percent is done, need to be non-zero
        if seed == None:
            self.seed = -1                # non-negative int, seed for random number generation
        else:
            self.seed = seed
        self.check_overflow = check_overflow

        if check_overflow:
            check_overflow_func(self)

        self.init_storage()             # initialize storage bins. Put in a separate function because might want to change maxtime 
                                        # and that doesn't need to initialze the whole object again


    def init_storage(self):
        # initialize storage bins
        self.data_empty = True      # whether data storage bins are empty. model.run will refuse to run (raise error) if not empty.
        self.max_record = int(self.maxtime / self.record_itv)   # int, how many data points to store sin total
        self.compress_itv = 1       # int, intended to reduce size of data (if not 1). Updated by compress_data function
                                    # if set to an int, say 20, mod will take average over every 20 data points and save them as new data.
                                    # May be used over and over again to recursively reduce data size. 
                                    # Default is 1, not to take average.
        self.U = None               # initialized by simulation.run or data_tools.read_data
        self.V = None
        self.Upi = None
        self.Vpi = None


    def check_valid_input(self, N, M, maxtime, record_itv, sim_time, boundary, I, X, P, print_pct, seed, check_overflow):
        # check whether the inputs are valid

        if (N < 1) or (M < 1):
            raise ValueError('N < 1 or M < 1')
        if (N == 1) and (M == 1):
            raise ValueError('Model fails for 1x1 space')
        if (M == 1):
            raise ValueError('Please set N = 1 for 1D space.')
        if maxtime <= 0:
            raise ValueError('Please set a positive number for maxtime')
        if record_itv <= 0:
            raise ValueError('Please set a positive number for record_itv')
        if sim_time <= 0:
            raise ValueError('Please set a positive number for sim_time')
        if type(boundary) != bool:
            raise TypeError('boundary not a bool. Please use True for zero-flux (with boundary) or False for periodical (no boundary)')
        
        if (type(I) != list) and (type(I) != np.ndarray):
            raise TypeError('Please set I as a list or np.ndarray')
        if np.array(I).shape != (N, M, 2):
            raise ValueError('Please set I as a N x M x 2 shape list or array. 2 is for init values of U, V at every patch')
        
        if (type(X) != list) and (type(X) != np.ndarray):
            raise TypeError('Please set X as a list or np.ndarray')
        if np.array(X).shape != (N, M, 4):
            raise ValueError('Please set X as a N x M x 4 shape list or array. 4 is for the flattened 2x2 payoff matrix')
        
        if (type(P) != list) and (type(P) != np.ndarray):
            raise TypeError('Please set P as a list or np.ndarray')
        if np.array(P).shape != (N, M, 6):
            raise ValueError('Please set P as a N x M x 6 shape list or array. 6 is for mu1, mu2, w1, w2, kappa1, kappa2')
        
        if not ((print_pct == None) or (isinstance(print_pct, int) and (print_pct >= -1))):
            # if not the two acceptable values: None or some >= -1 int
            raise ValueError('Please use an int > 0 for print_pct or None for not printing progress.')
        
        if not ((seed == None) or (isinstance(seed, int) and (seed >= -1))):
            raise ValueError('Please use a non-negative int as seed, or use None for no seed.')
        
        if not isinstance(check_overflow, bool):
            raise ValueError('Please use a bool for check_overflow')
        

    def check_valid_data(self, data_empty, max_record, compress_itv):
        # check whether a set of data is valid, used when reading a saved model
        if type(data_empty) != bool:
            raise TypeError('data_empty not a bool')
        
        if type(max_record) != int:
            raise TypeError('max_record not an int')
        if max_record < 0:
            raise ValueError('max_record < 0')
        
        if type(compress_itv) != int:
            raise TypeError('compress_itv not an int')
        if compress_itv < 0:
            raise ValueError('compress_itv < 0')
   

    def __str__(self):
        # print this mod in a nice format

        self_str = ''
        self_str += 'N = ' + str(self.N) + '\n'
        self_str += 'M = ' + str(self.M) + '\n'
        self_str += 'maxtime = ' + str(self.maxtime) + '\n'
        self_str += 'record_itv = ' + str(self.record_itv) + '\n'
        self_str += 'sim_time = ' + str(self.sim_time) + '\n'
        self_str += 'boundary = ' + str(self.boundary) + '\n'
        self_str += 'print_pct = ' + str(self.print_pct) + '\n'
        self_str += 'seed = ' + str(self.seed) + '\n'
        self_str += 'check_overflow = ' + str(self.check_overflow) + '\n'
        self_str += 'data_empty = ' + str(self.data_empty) + '\n'
        self_str += 'compress_itv = ' + str(self.compress_itv) + '\n'
        self_str += '\n'

        # check whether I, X, P all same (compare all patches to (0, 0))
        I_same = True
        X_same = True
        P_same = True
        for i in range(self.N):
            for j in range(self.M):
                for k in range(2):
                    if self.I[i][j][k] != self.I[0][0][k]:
                        I_same = False
                for k in range(4):
                    if self.X[i][j][k] != self.X[0][0][k]:
                        X_same = False
                for k in range(6):
                    if self.P[i][j][k] != self.P[0][0][k]:
                        P_same = False
        
        if I_same:
            self_str += 'I all same: ' + str(self.I[0][0]) + '\n'
        else:
            self_str += 'I:\n'
            for i in range(self.N):
                for j in range(self.M):
                    self_str += str(self.I[i][j]) + ' '
                self_str += '\n'
            self_str += '\n'
        
        if X_same:
            self_str += 'X all same: ' + str(self.X[0][0]) + '\n'
        else:
            self_str += 'X:\n'
            for i in range(self.N):
                for j in range(self.M):
                    self_str += str(self.X[i][j]) + ' '
                self_str += '\n'
            self_str += '\n'

        if P_same:
            self_str += 'P all same: ' + str(self.P[0][0]) + '\n'
        else:
            self_str += 'P:\n'
            for i in range(self.N):
                for j in range(self.M):
                    self_str += str(self.P[i][j]) + ' '
                self_str += '\n'

        return self_str
    

    def copy(self, copy_data = True):
        # return deep copy of self
        # copy_data decides whether to copy data as well
        if type(copy_data) != bool:
            raise TypeError('Please give a bool as argument: whether to copy data or not')

        sim2 = model(N = self.N, M = self.M, maxtime = self.maxtime, record_itv = self.record_itv, sim_time = self.sim_time, boundary = self.boundary,
                          I = np.copy(self.I), X = np.copy(self.X), P = np.copy(self.P), 
                          print_pct = self.print_pct, seed = self.seed, check_overflow = self.check_overflow)

        if copy_data:
            # copy data as well
            sim2.set_data(self.data_empty, self.max_record, self.compress_itv, self.U, self.V, self.Upi, self.Vpi)

        return sim2


    def calculate_ave(self):
        # get the average value over sim_time many models
        if self.sim_time != 1:
            for i in range(self.N):
                for j in range(self.M):
                    for t in range(self.max_record):
                        self.U[i][j][t] /= self.sim_time
                        self.V[i][j][t] /= self.sim_time
                        self.Upi[i][j][t] /= self.sim_time
                        self.Vpi[i][j][t] /= self.sim_time


    def change_maxtime(self, maxtime):
        # change maxtime
        if (type(maxtime) != float) and (type(maxtime) != int):
            raise TypeError('Please pass in a float or int as the new maxtime.')
        if maxtime <= 0:
            raise ValueError('Please use a positive maxtime.')
        self.maxtime = maxtime
        self.init_storage()


    def set_seed(self, seed):
        # set seed
        self.seed = seed


    def clear_data(self):
        # clear all data stored, set all to 0
        self.init_storage()


    def set_data(self, data_empty, max_record, compress_itv, U, V, Upi, Vpi):
        # set data to the given data values
        # copies are made
        self.check_valid_data(data_empty, max_record, compress_itv)

        self.data_empty = data_empty
        self.max_record = max_record
        self.compress_itv = compress_itv
        self.U = np.copy(U)
        self.V = np.copy(V)
        self.Upi = np.copy(Upi)
        self.Vpi = np.copy(Vpi)


    def compress_data(self, compress_itv = 5):
        # compress data by only storing average values
        if self.data_empty:
            raise RuntimeError('Model has empty data. Cannot compress')

        if type(compress_itv) != int:
            raise TypeError('Please use an int as compress_itv')
        if compress_itv < 1:
            raise ValueError('Please use record_itv >= 1')
        if compress_itv == 1:
            return
        
        self.compress_itv *= compress_itv  # may be reduced over and over again
        self.max_record = int(self.max_record / compress_itv)  # number of data points after reducing

        U_reduced = np.zeros((self.N, self.M, self.max_record), dtype = np.float64)
        V_reduced = np.zeros((self.N, self.M, self.max_record), dtype = np.float64)
        Upi_reduced = np.zeros((self.N, self.M, self.max_record), dtype = np.float64)
        Vpi_reduced = np.zeros((self.N, self.M, self.max_record), dtype = np.float64)

        for i in range(self.N):
            for j in range(self.M):
                for k in range(self.max_record):
                    lower = k * compress_itv  # upper and lower bound of current record_itv
                    upper = lower + compress_itv
                    U_reduced[i][j][k] = np.mean(self.U[i, j, lower : upper])
                    V_reduced[i][j][k] = np.mean(self.V[i, j, lower : upper])
                    Upi_reduced[i][j][k] = np.mean(self.Upi[i, j, lower : upper])
                    Vpi_reduced[i][j][k] = np.mean(self.Vpi[i, j, lower : upper])
        
        self.U = U_reduced
        self.V = V_reduced
        self.Upi = Upi_reduced
        self.Vpi = Vpi_reduced



def run(mod, message = "", update_sum_freq = UPDATE_SUM_FREQ):
    '''
    C-cored simulation
    '''

    read_lib()

    if not mod.data_empty:
        raise ValueError('mod has non-empty data.')

    msg_len = len(message) * 2
    msg_bytes = message.encode('utf-8')
    msg_buffer = ctypes.create_string_buffer(msg_bytes, msg_len) 
    
    I = np.ascontiguousarray(mod.I.flatten(), dtype = np.uint32)
    X = np.ascontiguousarray(mod.X.flatten(), dtype = np.float64)
    P = np.ascontiguousarray(mod.P.flatten(), dtype = np.float64)

    mod_c = model_c()
    success = LIB.mod_init(ctypes.byref(mod_c), 
                           mod.N, mod.M, mod.maxtime, mod.record_itv, mod.sim_time, mod.boundary,
                           I, X, P, mod.print_pct, mod.seed)
    if not success:
        LIB.mod_free_py(ctypes.byref(mod_c))
        del mod_c
        raise RuntimeError('Model initialization failed')
    
    result = LIB.run(ctypes.byref(mod_c), msg_buffer, msg_len, update_sum_freq)

    if result == SIM_SUCCESS:
        mod.set_data(False, mod.max_record, 1, mod_c.get_array('U1d').reshape(mod.N, mod.M, mod.max_record), 
                    mod_c.get_array('V1d').reshape(mod.N, mod.M, mod.max_record), 
                    mod_c.get_array('Upi_1d').reshape(mod.N, mod.M, mod.max_record), 
                    mod_c.get_array('Vpi_1d').reshape(mod.N, mod.M, mod.max_record))
        LIB.mod_free_py(ctypes.byref(mod_c))
        del mod_c
    elif result == SIM_SMALL_MAXTIME:
        LIB.mod_free_py(ctypes.byref(mod_c))
        del mod_c
        raise RuntimeError('maxtime too small.')
    elif result == SIM_OVERFLOW:
        LIB.mod_free_py(ctypes.byref(mod_c))
        del mod_c
        raise OverflowError('Overflow in simulation')
    else:
        LIB.mod_free_py(ctypes.byref(mod_c))
        del mod_c
        raise RuntimeError('Unclassified error.')




def demo_model():
    '''
    Returns a demo model.model object
    '''

    N = 10                  # Number of rows
    M = 10                  # Number of cols
    maxtime = 30           # how long you want the model to run
    record_itv = 0.1        # how often to record data.
    sim_time = 1            # repeat model to reduce randomness
    boundary = True         # boundary condition.

    # initial population for the N x M patches. 
    I = [[[44, 22] for _ in range(M)] for _ in range(N)]
    
    # flattened payoff matrices, total resource is 0.4, cost of fighting is 0.1
    X = [[[-1, 4, 0, 2] for _ in range(M)] for _ in range(N)]
    
    # patch variables
    P = [[[0.5, 0.5, 1, 1, 0.001, 0.001] for _ in range(M)] for _ in range(N)]

    print_pct = 25           # print progress
    seed = 36               # seed for random number generation

    # create a model object
    mod = model(N, M, maxtime, record_itv, sim_time, boundary, I, X, P, 
                            print_pct = print_pct, seed = seed)
    
    return mod



def UV_expected_val(mod):
    '''
    Calculate expected U & V population and payoff based on matrices, assume no migration and any stochastic process.
    To differentiate from UV_expected in figures.py: this one return arrays (values).
    '''
    
    U_expected = np.zeros((mod.N, mod.M))  # expected U population
    V_expected = np.zeros((mod.N, mod.M))  # expected V population
    pi_expected = np.zeros((mod.N, mod.M)) # expected payoff, which are equal for U and V
    
    for i in range(mod.N):
        for j in range(mod.M):
            # say matrix = [a, b, c, d]
            # U_proportion = (d - b) / (a - b - c + d)
            U_prop = (mod.X[i][j][3] - mod.X[i][j][1]) / (mod.X[i][j][0] - mod.X[i][j][1] - mod.X[i][j][2] + mod.X[i][j][3])
            # equilibrium payoff, U_payoff = V_payoff
            eq_payoff = U_prop * mod.X[i][j][0] + (1 - U_prop) * mod.X[i][j][1]
            
            # payoff / kappa * proportion
            U_expected[i][j] = eq_payoff / mod.P[i][j][4] * U_prop
            V_expected[i][j] = eq_payoff / mod.P[i][j][5] * (1 - U_prop)
            pi_expected[i][j] = eq_payoff
                
    return U_expected, V_expected, pi_expected



def check_overflow_func(mod):
    _, _, pi_expected = UV_expected_val(mod)
    for i in range(mod.N):
        for j in range(mod.M):
            w1_pi = pi_expected[i][j] * mod.P[i][j][2]  # w1 * U_pi
            w2_pi = pi_expected[i][j] * mod.P[i][j][3]  # w2 * V_pi
            if ((w1_pi > EXP_OVERFLOW_BOUND) or (w2_pi > EXP_OVERFLOW_BOUND)):
                print("Warning: might cause overflow. \n\t w1, w2, or payoff matrix values too large")
                return
            if ((w1_pi > EXP_TOO_LARGE_BOUND) or (w2_pi > EXP_TOO_LARGE_BOUND)):
                print("Warning: might encounter large values > 3e38 in simulation. \n\t w1, w2, or payoff matrix values too large")
                return

